@@ -20,6 +20,10 @@ from webob.statusreasons import status_reasons
 from webob.cachecontrol import CacheControl, serialize_cache_control
 from webob.acceptparse import Accept, MIMEAccept, NilAccept, MIMENilAccept, NoAccept
 from webob.byterange import Range, ContentRange
+try:
+    sorted
+except NameError:
+    from webob.compat import sorted
 
 _CHARSET_RE = re.compile(r';\s*charset=([^;]*)', re.I)
 _SCHEME_RE = re.compile(r'^[a-z]+:', re.I)
@@ -301,6 +305,17 @@ def _serialize_date(dt):
             "You must pass in a datetime, date, time tuple, or integer object, not %r" % dt)
     return formatdate(dt)
 
+def _serialize_cookie_date(dt):
+    if dt is None:
+        return None
+    if isinstance(dt, unicode):
+        dt = dt.encode('ascii')
+    if isinstance(dt, timedelta):
+        dt = datetime.now() + dt
+    if isinstance(dt, (datetime, date)):
+        dt = dt.timetuple()
+    return time.strftime('%a, %d-%b-%Y %H:%M:%S GMT', dt)
+
 def _parse_date_delta(value):
     """
     like _parse_date, but also handle delta seconds
@@ -468,7 +483,7 @@ class Request(object):
     request_body_tempfile_limit = 10*1024
 
     def __init__(self, environ=None, environ_getter=None, charset=NoDefault, unicode_errors=NoDefault,
-                 decode_param_names=NoDefault):
+                 decode_param_names=NoDefault, **kw):
         if environ is None and environ_getter is None:
             raise TypeError(
                 "You must provide one of environ or environ_getter")
@@ -488,6 +503,11 @@ class Request(object):
             self.__dict__['unicode_errors'] = unicode_errors
         if decode_param_names is not NoDefault:
             self.__dict__['decode_param_names'] = decode_param_names
+        for name, value in kw.items():
+            if not hasattr(self.__class__, name):
+                raise TypeError(
+                    "Unexpected keyword: %s=%r" % name, value)
+            setattr(self, name, value)
 
     def __setattr__(self, attr, value, DEFAULT=[]):
         ## FIXME: I don't know why I need this guard (though experimentation says I do)
@@ -1010,16 +1030,9 @@ class Request(object):
         This only does a shallow copy, except of wsgi.input
         """
         env = self.environ.copy()
-        data = self.body
-        tempfile_limit = self.request_body_tempfile_limit
-        if tempfile_limit and len(data) > tempfile_limit:
-            fileobj = tempfile.TemporaryFile()
-            fileobj.write(data)
-            fileobj.seek(0)
-        else:
-            fileobj = StringIO(data)
-        env['wsgi.input'] = fileobj
-        return self.__class__(env)
+        new_req = self.__class__(env)
+        new_req.copy_body()
+        return new_req
 
     def copy_get(self):
         """
@@ -1035,7 +1048,64 @@ class Request(object):
         env['REQUEST_METHOD'] = 'GET'
         return self.__class__(env)
 
-    def remove_conditional_headers(self, remove_encoding=True):
+    def make_body_seekable(self):
+        """
+        This forces ``environ['wsgi.input']`` to be seekable.  That
+        is, if it doesn't have a `seek` method already, the content is
+        copied into a StringIO or temporary file.
+
+        The choice to copy to StringIO is made from
+        ``self.request_body_tempfile_limit``
+        """
+        input = self.body_file
+        if hasattr(input, 'seek'):
+            # It has a seek method, so we don't need to do anything
+            return
+        self.copy_body()
+
+    def copy_body(self):
+        """
+        Copies the body, in cases where it might be shared with
+        another request object and that is not desired.
+
+        This copies the body in-place, either into a StringIO object
+        or a temporary file.
+        """
+        length = self.content_length
+        if length == 0:
+            # No real need to copy this, but of course it is free
+            self.body_file = StringIO('')
+            return
+        tempfile_limit = self.request_body_tempfile_limit
+        body = None
+        input = self.body_file
+        if hasattr(input, 'seek'):
+            # Just in case someone has read parts of the body already
+            ## FIXME: Should we use .tell() to try to put the body
+            ## back to its previous position?
+            input.seek(0)
+        if length == -1:
+            body = self.body
+            length = len(body)
+            self.content_length = length
+        if tempfile_limit and length > tempfile_limit:
+            fileobj = tempfile.TemporaryFile()
+            if body is None:
+                while length:
+                    data = input.read(min(length, 4096))
+                    fileobj.write(data)
+                    length -= len(data)
+            else:
+                fileobj.write(body)
+            fileobj.seek(0)
+        else:
+            if body is None:
+                body = input.read(length)
+            fileobj = StringIO(body)
+        self.body_file = fileobj
+
+    def remove_conditional_headers(self, remove_encoding=True, remove_range=True,
+                                        remove_match=True, remove_modified=True):
         """
         Remove headers that make the request conditional.
 
@@ -1045,13 +1115,19 @@ class Request(object):
         This does not remove headers like If-Match, which are used for
         conflict detection.
         """
-        for key in ['HTTP_IF_MATCH', 'HTTP_IF_MODIFIED_SINCE',
-                    'HTTP_IF_RANGE', 'HTTP_RANGE']:
+        check_keys = []
+        if remove_range:
+            check_keys += ['HTTP_IF_RANGE', 'HTTP_RANGE']
+        if remove_match:
+            check_keys.append('HTTP_IF_MATCH')
+        if remove_modified:
+            check_keys.append('HTTP_IF_MODIFIED_SINCE')
+        if remove_encoding:
+            check_keys.append('HTTP_ACCEPT_ENCODING')
+
+        for key in check_keys:
             if key in self.environ:
                 del self.environ[key]
-        if remove_encoding:
-            if 'HTTP_ACCEPT_ENCODING' in self.environ:
-                del self.environ['HTTP_ACCEPT_ENCODING']
 
     accept = converter(
         environ_getter('HTTP_ACCEPT', rfc_section='14.1'),
@@ -1385,7 +1461,7 @@ class Response(object):
             if not hasattr(self.__class__, name):
                 # Not a basic attribute
                 raise TypeError(
-                    "Unexpected keyword: %s=%r in %r" % (name, value))
+                    "Unexpected keyword: %s=%r" % (name, value))
             setattr(self, name, value)
 
     def __repr__(self):
@@ -1696,7 +1772,7 @@ class Response(object):
         if isinstance(expires, timedelta):
             expires = datetime.utcnow() + expires
         if isinstance(expires, datetime):
-            expires = '"'+_serialize_date(expires)+'"'
+            expires = '"'+_serialize_cookie_date(expires)+'"'
         for var_name, var_value in [
             ('max_age', max_age),
             ('path', path),
@@ -1936,18 +2012,22 @@ class Response(object):
 
     etag = header_getter('ETag', rfc_section='14.19')
 
-    def md5_etag(self, body=None):
+    def md5_etag(self, body=None, set_content_md5=False):
         """
         Generate an etag for the response object using an MD5 hash of
         the body (the body parameter, or ``self.body`` if not given)
 
         Sets ``self.etag``
+        If set_content_md5 is True sets self.content_md5 as well
         """
         if body is None:
             body = self.body
         import md5
         h = md5.new(body)
-        self.etag = h.digest().encode('base64').replace('\n', '').strip('=')
+        md5_digest = h.digest().encode('base64').replace('\n', '').strip('=')
+        self.etag = md5_digest
+        if set_content_md5:
+            self.content_md5 = md5_digest
 
     expires = converter(
         header_getter('Expires', rfc_section='14.21'),
